=begin

Copyright 2011, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         LaunchUp.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'LaunchUp - Tool Launcher'
              or create a shortcut, for example ctrl+space or ctrl+f
Description:  This Plugin adds a quick launcher to
                (1) Load plugins on the fly (if some plugins are not yet loaded or disabled)
                    requirement: index must contain a name and the path to the ruby file
                    Then the plugin is available at least from the menu/toolbars like normal, 
                    or it will be directly executed by (2).
                (2) Execute tools. 
                    requirement: (1), index must also contain the command of the tool
                    If necessary, the required plugin is also loaded on the fly.
Version:      0.2.1  
Date:         25.09.2011

=end



module LaunchUp
require 'sketchup.rb'
@dir = File.dirname(__FILE__) # this plugin's folder
@index_file = File.join(@dir,"index.txt") # index file
@plugins_dir = Sketchup.find_support_file("Plugins") # TODO: what about (more than one) loadpaths?
@windowTitlebarHeight = 0
@windowBorderWidth = 0

@launchdlg = nil #FIXME: remove this after debugging


##
# main method, displays webdialog
#
def self.run
  if !@launchdlg
    # read index file
    return UI.messagebox("LaunchUp error: The index file could not be found.", MB_OK) if !File.exists?(@index_file)
    @index = IO.read(@index_file) # string of the index
    return (@index="[]";UI.messagebox("LaunchUp error: The index file is empty.", MB_OK)) if @index.empty?
    # create webdialog
    #@launchdlg = UI::WebDialog.new("LaunchUp", false, false, 0, 1140, 270, 50, true)
    @launchdlg = UI::WebDialog.new("LaunchUp", false, false, 800, 200, 270, 50, true)
    html_path = File.join(@dir, "LaunchUp.htm")
    @launchdlg.set_file(html_path)
    @launchdlg.set_size(270,50)
    @launchdlg.add_action_callback("onload") {|@launchdlg, param|
      @launchdlg.execute_script("index="+@index)
      @launchdlg.execute_script("init()")
    }
    @launchdlg.add_action_callback("launch") {|@launchdlg, param|
      #eval(@launchdlg.get_element_value("return-value"))
      begin 
        params = eval(@launchdlg.get_element_value("return-value"))
        if !params["files"].empty? && !params["files"][0].empty?
          begin 
            params["files"].each{|f|
              (load(f); puts "load "+f ) if !file_loaded?(f) && FileTest.file?(File.join(@plugins_dir,f))
              (load(f+"!"); puts "load "+f+"!" ) if !file_loaded?(f+"!") && FileTest.file?(File.join(@plugins_dir,f+"!")) # disabled plugin
              (load(f+".txt"); puts "load "+f+".txt" ) if !file_loaded?(f+".txt") && FileTest.file?(File.join(@plugins_dir,f+"txt")) # disabled plugin
            }
            eval(params["action"])
          rescue ScriptError
            puts "LaunchUp Error: loading #{params["files"]}"
          end
        end
      rescue
        puts "LaunchUp Error: webdialog returned invalid data"
      end
    }
    @launchdlg.add_action_callback("windowInner") {|@launchdlg, param|
      param = eval(param)
      @windowBorderWidth = (270-param[0].to_f)/2 if @windowBorderWidth == 0
      @windowTitlebarHeight = 50-param[1].to_f if @windowTitlebarHeight == 0
    }
    @launchdlg.add_action_callback("resizeHeight") {|@launchdlg, param|
      param = eval(param)
      width = 2*@windowBorderWidth+param[0].to_f; width = 270 if width<270
      height = @windowTitlebarHeight+param[1].to_f; height = 50 if height<50
      @launchdlg.set_size(width, height)
    }
    # for debugging
    @launchdlg.add_action_callback("puts") {|@launchdlg, param|
      puts param.to_s
    }
    if RUBY_PLATFORM.include?('darwin') # Do NOT do this if on a PC!
      @launchdlg.show_modal
    else
      @launchdlg.show
    end #if mac _modal
  # toggle webdialog. Useful for shortcuts
  elsif @launchdlg.visible?
    @launchdlg.bring_to_front
    @launchdlg.execute_script("$('combo-input').focus()")
    # @launchdlg.close
  else
    if RUBY_PLATFORM.include?('darwin') # Do NOT do this if on a PC!
      @launchdlg.show_modal
    else
      @launchdlg.show
    end #if mac _modal
  end
end #def run



##
# *Experimental*
#  Parse ruby files in order to extract if possible the name and command of tools
#
def self.build_index
@index=""
  # @index is a json string
  # index is a hash, currently only needed in this method
  index = eval(@index.gsub(":","=>")) || []
  d = @plugins_dir
  Dir.foreach(d){|f|
    path = File.join(d,f)
    # only continue with ruby files
    next if f == "."|| f == ".." || !FileTest.file?(path) || f[/\.rb(?:\!|\.txt)?$/].nil?
    relpath = self.rel_path(path, @plugins_dir).gsub(/\.rb(?:\!|\.txt)?$/,".rb")
    # skip if file is already in index
    next if index!=nil && nil!=index.detect{|i| i["file"]==relpath}
    item = self.parse_ruby(path)
    index.push(item)
  }
  # create again json string from new index
  @index = index.inspect.gsub("=>", ":")
  # update index in webdialog
  @launchdlg.execute_script("index="+@index) if @launchdlg!=nil
  # save new index to file
  # this index files needs manual cleanup (remove support files/libraries) and check if commands are complete and correct
  @out = File.new(@index_file, "w")
  @index = self.indent_json(@index)
  @out.puts @index
  @out.close
end #def build_index



def self.parse_ruby(path)
  relpath = self.rel_path(path, @plugins_dir).gsub(/\.rb(?:\!|\.txt)?$/,".rb")
  # use filename as preliminary tool name; clean it up
  name = relpath[/\/?.*?$/].gsub(/\.rb$|loader|ZLoader_*?/,'').gsub(/_/,' ').gsub(/([a-z])([A-Z])/m,'\1 \2')
  commands = []
  # ruby file is only a loader => find all referred ruby files and parse them instead
  if relpath[/loader/i]
    loaderContent = IO.read(path)
    supportFiles = loaderContent.scan(/(?:load|require)\s?\(?\s?(?:'|")(.*)(?:'|")\s?\)?/m)
    # loaderContent[/(?:load|require).*$/].scan(/(?:'|")(.*?)(?:'|")/) # File.join(*)
    supportFiles.each{|d|
      p = File.join(@plugins_dir, d)
      commands.concat(self.parse_commands(p)) if File.exists?(p)
    }
  # otherwise parse the single ruby file
  else
    commands.concat(self.parse_commands(path))
  end
  # create item for index
  item = {"name" => name, "file" => relpath}
  # plugins that contain more than one command (many tools)
  if commands.length > 1
    item["tools"] = []
    commands.each{|c|
      item["tools"].push({"name" => c[0], "cmd" => c[1]}) 
    }
  # plugins with only one command
  elsif commands.length == 1
    item["name"] = commands[0][0]
    item["cmd"] = commands[0][1]
  end
  return item
end #def parse_ruby



##
# Parse a single ruby file and return extracted commands
#
def self.parse_commands(path)
  content = IO.read(path)
  # extract commands from UI:Command.new or menu.add_item or toolbar.add_item...
  commands = content.scan(/UI\:\:Command.new\s*\(\s*(?:'|")\s*(.*?)\s*(?:'|")\s*\)\s*\{\s*(.*?(?:\{.*?\})?)\s*\}/)
  commands.concat(content.scan(/.add_item\s*\(\s*(?:'|")\s*(.*?)\s*(?:'|")\s*\)\s*\{\s*(.*?(?:\{.*?\})?)\s*\}/))
  return commands
end



########## helper methods ##########



##
# Convert absolute file path into relative path, relative to
# - <base> directory or
# - next directory that is common with <file> given as second argument
#
def self.rel_path(abs,base=@plugins_dir)
  base=base.sub(/\/?$/,"/")
  i=0
  i+=1 until abs[i]!=base[i]
  i-=1 if i>0
  rel = abs.scan(/^.{0,#{i}}\/(.*$)/).to_s
  return (i==0)? abs : rel
end #def rel_path



###
#  Format a json string so that it is easier to read
#
# translated to ruby from http://ketanjetty.com/coldfusion/javascript/format-json/
#
def self.indent_json(json)
  result = ''
  pos = 0
  strLen = json.length
  indentStr = ' '
  indentLen = 2
  newLine = "\n"
  prevChar = ''
  outOfQuotes = true

  for i in 0..strLen
    # Grab the next character in the string.
    char = json[i,1]
    # Are we inside a quoted string?
    if (char == '"' && prevChar != '\\')
        outOfQuotes = !outOfQuotes
    # If this character is the end of an element, 
    # output a new line and indent the next line.
    elsif ((char == '}' || char == ']') && outOfQuotes)
        result += newLine
        pos -= 1
        result += indentStr*indentLen*pos
    elsif (char == " " && prevChar != "\"" && prevChar != ":" && outOfQuotes)
      next
    end
    # Add the character to the result string.
    result += char
    # If the last character was the beginning of an element, 
    # output a new line and indent the next line.
    if ((char == ',' || char == '{' || char == '[') && outOfQuotes)
        result += newLine
        if (char == '{' || char == '[')
            pos += 1
        end
        result += indentStr*indentLen*pos
    end
    prevChar = char
  end
  return result
end #def indent_json



end #module LaunchUp



if( not file_loaded?(File.basename(__FILE__)) )
  cmd_launchup = UI::Command.new("LaunchUp - Tool Launcher"){LaunchUp.run}
	UI.menu("Plugins").add_item(cmd_launchup)
end

file_loaded(File.basename(__FILE__))
